from typing import List

import numpy as np

EPISILON = 1.48e-8


def diff_3point_inplace(
    func: callable, variable_col_idx_list: List[int], residual_row_start: int, jac: np.ndarray, *variables
):
    """3-point numeric finite difference. f'(x) = (f(x+h) - f(x-h)) / 2h
    :param jac_shape: tuple
        the shape of returned jacobian matrix.
    :param func: callable
        the function to evaluate jacobian matrix.
    :param variables: ndarrays
        At which the jacobian matrix is evaluated.
    :return:
    """
    for col, variable in zip(variable_col_idx_list, variables):
        h = np.maximum(EPISILON * variable, EPISILON)
        for i in range(h.shape[0]):
            variable[i] += h[i]
            f_plus = func(*variables)
            variable[i] -= 2 * h[i]
            f_subs = func(*variables)
            slope = (f_plus - f_subs) / (2 * h[i])
            jac[residual_row_start : residual_row_start + slope.size, col + i] = slope
            variable[i] += h[i]



from py_minisam.optimizer.gauss_newton_optimizer import GaussNewtonOptimizer
from py_minisam.optimizer.levenberg_marquardt_optimizer import LevenbergMarquardtOptimizer

__all__ = ["GaussNewtonOptimizer", "LevenbergMarquardtOptimizer"]



import numpy as np

from py_minisam.optimizer.base_optimizer import BaseOptimizer, ProblemResult, SolverParameters, SolverStatus
from py_minisam.problem import Problem


class LevenbergMarquardtOptimizer(BaseOptimizer):
    def __init__(self) -> None:
        pass

    def optimize(self, problem: Problem, max_iteration: int = 100):
        result = ProblemResult()
        solver_params = SolverParameters()
        v = 2
        u = 0.0

        for i in range(max_iteration):
            result.iterations += 1

            params = problem.combine_variables()
            residuals, jac = problem.compute_residual_and_jacobian(params)
            gradient = jac.T @ -residuals
            jtj = jac.T @ jac
            max_gradient = np.amax(np.abs(gradient))
            if max_gradient < solver_params.gradient_threshold:
                print("g small")
                break
            elif np.linalg.norm(residuals) < solver_params.error_threshold:
                print("err small")
                break
            if i == 0:
                u = solver_params.initial_scale_factor * np.amax(np.diag(jtj))

            jtj_augmented = jtj.copy()
            np.fill_diagonal(jtj_augmented, np.diag(jtj_augmented) + u)
            dx = np.linalg.solve(jtj_augmented, gradient)
            solution = jtj_augmented @ dx
            if np.amin(np.abs(solution - gradient)) < solver_params.error_threshold:
                if np.linalg.norm(dx) < solver_params.relative_step_threshold * np.linalg.norm(params):
                    result.status = SolverStatus.RelativeStepSizeTooSmall
                    break
                param_new = params + dx
                residual_new = np.zeros(problem._dim_residual, np.float64)
                for residual_block in problem.residual_blocks:
                    variables = []
                    for col in residual_block.variable_col_start_index_list:
                        variables.append(problem.col_idx_to_variable_dict[col])
                    residual = residual_block.residual_func(*variables)
                    residuals[
                        residual_block.residual_row_start_idx : residual_block.residual_row_start_idx + residual.size
                    ] = residual
                r_norm = np.linalg.norm(residual)
                r_norm_new = np.linalg.norm(residual_new)

                rho = (r_norm * r_norm - r_norm_new * r_norm_new) / np.dot(dx, (u * dx + gradient))
                if rho > 0.0:
                    # params = param_new
                    print("good")
                    problem.write_back_variables(param_new)
                    tmp = 2.0 * rho - 1.0
                    u = u * max(1.0 / 3.0, 1.0 - tmp**3)
                    v = 2
                    continue
            else:
                result.num_failed_linear_solves += 1
                print(f"fail {solution - gradient}")
            u *= v
            v *= 2



from dataclasses import dataclass
from functools import partial
from typing import Dict, List, Tuple

import numpy as np

from py_minisam.auto_diff import diff_3point_inplace
from py_minisam.factor_graph import FactorGraph


@dataclass
class ResidualBlock:

    dim_residual: int
    residual_row_start_idx: int
    variable_col_start_index_list: List[int]
    residual_func: callable
    jac_func = None
    loss_func = None
    jac_sparsity = None

    def __post_init__(self):
        self.jac_func = partial(
            diff_3point_inplace, self.residual_func, self.variable_col_start_index_list, self.residual_row_start_idx
        )

    def jacobians_inplace(self, jacobian_matrix: np.ndarray, *variables):
        self.jac_func(jacobian_matrix, *variables)


class Problem:
    def __init__(self) -> None:
        self._dim_variable = 0
        self._dim_residual = 0
        self.residual_blocks: List[ResidualBlock] = []
        self.variable_addr_to_col_idx_dict: Dict[int, int] = {}
        self.col_idx_to_variable_dict: Dict[int, np.ndarray] = {}

    @staticmethod
    def from_factor_graph(factor_graph: FactorGraph, initial_variables: Dict[str, np.ndarray]):
        problem = Problem()
        for factor in factor_graph.factors:
            variables = (initial_variables[k] for k in factor.variable_key_list)
            problem.add_residual_block(factor.dim_residual, factor.error_func, *variables)
        return problem

    def add_residual_block(self, dim_residual, residual_func, *variables):
        if dim_residual <= 0:
            err_msg = f"dim_residual should > 0, got {dim_residual}"
            raise ValueError(err_msg)
        if not callable(residual_func):
            err_msg = "residual_func should be callable"
            raise TypeError(err_msg)

        variable_col_start_index_list = []
        for variable in variables:
            if not isinstance(variable, np.ndarray):
                raise TypeError
            address = variable.__array_interface__["data"][0]
            if address not in self.variable_addr_to_col_idx_dict:
                variable_dim = variable.shape[0]
                self.variable_addr_to_col_idx_dict[address] = self._dim_variable
                self.col_idx_to_variable_dict[self._dim_variable] = variable
                self._dim_variable += variable_dim
            variable_col_start_index_list.append(self.variable_addr_to_col_idx_dict[address])

        residual_block = ResidualBlock(
            dim_residual=dim_residual,
            residual_row_start_idx=self._dim_residual,
            variable_col_start_index_list=variable_col_start_index_list,
            residual_func=residual_func,
        )
        self.residual_blocks.append(residual_block)
        self._dim_residual += dim_residual

    def compute_residual_and_jacobian(self, params: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        jac = np.zeros((self._dim_residual, self._dim_variable), dtype=np.float64)
        residuals = np.zeros(self._dim_residual, dtype=np.float64)
        for residual_block in self.residual_blocks:
            variables = []
            for col in residual_block.variable_col_start_index_list:
                variables.append(params[col : col + self.col_idx_to_variable_dict[col].size])
            residual = residual_block.residual_func(*variables)
            residuals[residual_block.residual_row_start_idx : residual_block.residual_row_start_idx + residual.size] = (
                residual
            )
            residual_block.jacobians_inplace(jac, *variables)
        return residuals, jac

    def combine_variables(self) -> np.ndarray:
        all_variables = np.zeros(self._dim_variable, np.float64)
        for col, variable in self.col_idx_to_variable_dict.items():
            all_variables[col : col + variable.size] = variable
        return all_variables

    def write_back_variables(self, all_variables: np.ndarray):
        for col, variable in self.col_idx_to_variable_dict.items():
            variable[:] = all_variables[col : col + variable.size]



import numpy as np

from py_minisam.optimizer.base_optimizer import BaseOptimizer, ProblemResult
from py_minisam.problem import Problem


class GaussNewtonOptimizer(BaseOptimizer):
    def __init__(self) -> None:
        pass

    def optimize(self, problem: Problem, max_iteration: int = 100):
        result = ProblemResult()
        params = problem.combine_variables()
        for i in range(max_iteration):
            print(i)
            result.iterations += 1

            residuals, jac = problem.compute_residual_and_jacobian(params)
            hessian = jac.T @ jac
            b = -jac.T @ residuals
            dx = np.linalg.solve(hessian, b)
            if np.linalg.norm(dx) < 1e-8:
                break
            params += dx
            problem.write_back_variables(params)
        print(np.linalg.norm(residuals))



from abc import abstractmethod
from typing import List

import numpy as np


class FactorBase:
    def __init__(
        self,
        dim_residual: int,
        variable_key_list: List[str],
    ) -> None:
        self.dim_residual = dim_residual
        self.variable_key_list = variable_key_list

    @abstractmethod
    def error_func(self, *variables) -> np.ndarray:
        pass


class FactorGraph:
    def __init__(self) -> None:
        self.factors: List[FactorBase] = []

    def add(self, factor: FactorBase):
        if not isinstance(factor, FactorBase):
            raise ValueError
        self.factors.append(factor)



from functools import partial
from time import perf_counter

import numpy as np
import pytest
from scipy.spatial.transform import Rotation

from py_minisam.optimizer import GaussNewtonOptimizer, LevenbergMarquardtOptimizer
from py_minisam.problem import Problem


def cost_function(rmat_gt: np.ndarray, p3d: np.ndarray, rvec: np.ndarray):
    rotation_matrix = Rotation.from_rotvec(rvec).as_matrix()
    rotated_3d_points = rmat_gt @ p3d
    return ((rotation_matrix @ p3d) - rotated_3d_points).T.flatten()


@pytest.mark.parametrize("solver_class", [GaussNewtonOptimizer, LevenbergMarquardtOptimizer])
def test_optimizer_solve_rotation(solver_class):
    np.random.seed(2)
    rvec_gt = np.random.random(3)
    rmat_gt = Rotation.from_rotvec(rvec_gt).as_matrix()
    rvec_noise = rvec_gt + np.random.random(3) / 10.0

    rvec_noise_init = rvec_noise.copy()
    problem = Problem()

    p3d0 = np.random.random((3, 2))
    problem.add_residual_block(p3d0.size, partial(cost_function, rmat_gt, p3d0), rvec_noise_init)
    p3d1 = np.random.random((3, 3))
    problem.add_residual_block(p3d1.size, partial(cost_function, rmat_gt, p3d1), rvec_noise_init)
    solver = solver_class()
    solver.optimize(problem)
    assert np.allclose(rvec_gt, rvec_noise_init)



from functools import partial
from time import perf_counter
from typing import List

import numpy as np
from scipy.spatial.transform import Rotation

from py_minisam.factor_graph import FactorBase, FactorGraph
from py_minisam.optimizer import GaussNewtonOptimizer, LevenbergMarquardtOptimizer
from py_minisam.problem import Problem

np.random.seed(2)

rvec_gt = np.random.random(3)
rmat_gt = Rotation.from_rotvec(rvec_gt).as_matrix()


def cost_function(p3d: np.ndarray, rvec: np.ndarray):
    rotation_matrix = Rotation.from_rotvec(rvec).as_matrix()
    rotated_3d_points = rmat_gt @ p3d
    return ((rotation_matrix @ p3d) - rotated_3d_points).T.flatten()


class CustomFactor(FactorBase):
    def __init__(self, key: str, p3d: np.ndarray) -> None:
        self.p3d = p3d
        super().__init__(p3d.size, [key])

    def error_func(self, rvec) -> np.ndarray:
        rotation_matrix = Rotation.from_rotvec(rvec).as_matrix()
        rotated_3d_points = rmat_gt @ self.p3d
        return ((rotation_matrix @ self.p3d) - rotated_3d_points).T.flatten()


def main():
    rvec_noise = rvec_gt + np.random.random(3) / 10.0

    for solver_class in [GaussNewtonOptimizer, LevenbergMarquardtOptimizer]:
        rvec_noise_init = rvec_noise.copy()
        print("init:", rvec_noise_init)
        problem = Problem()

        p3d0 = np.random.random((3, 2))
        problem.add_residual_block(p3d0.size, partial(cost_function, p3d0), rvec_noise_init)
        p3d1 = np.random.random((3, 3))
        problem.add_residual_block(p3d1.size, partial(cost_function, p3d1), rvec_noise_init)
        print(f"{rvec_noise_init=}")
        solver = solver_class()
        solver.optimize(problem)
        print(f"{rvec_noise_init=}")
        print(f"{rvec_gt=}")

    # factor graph
    factor_graph = FactorGraph()
    p3d0 = np.random.random((3, 2))
    factor_graph.add(CustomFactor("rvec0", p3d0))
    p3d1 = np.random.random((3, 3))
    factor_graph.add(CustomFactor("rvec0", p3d1))
    initial_values = {"rvec0": rvec_noise.copy()}
    print(initial_values)
    gn = GaussNewtonOptimizer()
    gn.optimize_factor_graph(factor_graph, initial_values)
    print(initial_values)


if __name__ == "__main__":
    main()



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Dict

import numpy as np

from py_minisam.factor_graph import FactorGraph
from py_minisam.problem import Problem


class SolverStatus(Enum):
    Running = auto()
    GradientTooSmall = auto()  # eps > max(J'*f(x))
    RelativeStepSizeTooSmall = auto()  # eps > ||dx|| / ||x||
    ErrorTooSmall = auto()  # eps > ||f(x)||
    HitMaxIterations = auto()


@dataclass
class SolverParameters:
    gradient_threshold: float = 1e-16
    relative_step_threshold: float = 1e-16
    error_threshold: float = 1e-16
    initial_scale_factor: float = 1e-3
    max_iterations: int = 100


@dataclass
class ProblemResult:
    error_magnitude: float = 0.0
    gradient_magnitude: float = 0.0
    num_failed_linear_solves: int = 0
    iterations: int = 0
    status: SolverStatus = SolverStatus.Running


class BaseOptimizer:
    @abstractmethod
    def optimize(self, problem: Problem, max_iteration: int = 100):
        pass

    def optimize_factor_graph(self, factor_graph: FactorGraph, init_values: Dict[str, np.ndarray]):
        problem = Problem.from_factor_graph(factor_graph, init_values)
        self.optimize(problem)
